<?php

/**
 * i-doit
 *
 * @package    i-doit
 * @subpackage API
 * @author     Selcuk Kekec <skekec@i-doit.de>
 * @version    1.10
 * @copyright  synetics GmbH
 * @license    http://www.i-doit.com/license
 */

namespace idoit\Module\Api\Session;

use idoit\Module\Api\Session\SessionInterface;
use isys_component_session;

/**
 * Class UserSession
 *
 * @package idoit\Module\Api
 */
class UserSession implements SessionInterface
{
    /**
     * Username
     *
     * @var string
     */
    protected $username;

    /**
     * Password
     *
     * @var string
     */
    protected $password;

    /**
     * ApiKey
     *
     * @var string
     */
    protected $apiKey;

    /**
     * Persistent session
     *
     * @var bool
     */
    protected $persistent;

    /**
     * SessionComponent
     *
     * @var isys_component_session
     */
    protected $sessionComponent;

    /**
     * UserId
     *
     * @var int
     */
    protected $userId;

    /**
     * SessionId
     *
     * @var string
     */
    protected $sessionId;

    /**
     * MandatorId
     *
     * @var int
     */
    protected $mandatorId;

    /**
     * Check whether session exists
     *
     * @return bool
     */
    public function isLoggedIn()
    {
        return $this->getSessionComponent()
            ->is_logged_in();
    }

    /**
     * Refresh session
     *
     * @return bool
     * @throws \Exception
     */
    public function refreshSession()
    {
        // Get database component
        $databseComponent = \isys_application::instance()->container->get('database');

        // Check whether requirements are met
        if ($this->isLoggedIn() && is_object($databseComponent) && !empty($this->getSessionId())) {
            $sql = 'UPDATE isys_user_session SET
                    isys_user_session__time_last_action = CURRENT_TIMESTAMP,
                    isys_user_session__description      = \'' . $databseComponent->escape_string($_SERVER['REQUEST_URI']) . '\'
                    WHERE isys_user_session__php_sid    = \'' . $this->getSessionId() . '\'';

            // Refresh session by updating `last_action` column
            return $databseComponent->query($databseComponent->limit_update($sql, 1));
        }

        throw new \Exception('Unable to refresh session that does not exist.');
    }

    /**
     * @return string
     */
    public function getUsername()
    {
        return $this->username;
    }

    /**
     * @param string $username
     *
     * @return UserSession
     */
    public function setUsername($username)
    {
        $this->username = $username;

        return $this;
    }

    /**
     * @return string
     */
    public function getPassword()
    {
        return $this->password;
    }

    /**
     * @param string $password
     *
     * @return UserSession
     */
    public function setPassword($password)
    {
        $this->password = $password;

        return $this;
    }

    /**
     * @return string
     */
    public function getApiKey()
    {
        return $this->apiKey;
    }

    /**
     * @param string $apiKey
     *
     * @return UserSession
     */
    public function setApiKey($apiKey)
    {
        $this->apiKey = $apiKey;

        return $this;
    }

    /**
     * @return isys_component_session
     */
    public function getSessionComponent()
    {
        return $this->sessionComponent;
    }

    /**
     * @param isys_component_session $sessionComponent
     *
     * @return UserSession
     */
    public function setSessionComponent($sessionComponent)
    {
        $this->sessionComponent = $sessionComponent;

        return $this;
    }

    /**
     * @return bool
     */
    public function isPersistent()
    {
        return $this->persistent;
    }

    /**
     * @param bool $persistent
     *
     * @return UserSession
     */
    public function setPersistent($persistent)
    {
        $this->persistent = $persistent;

        return $this;
    }

    /**
     * @return int
     */
    public function getMandatorId()
    {
        return $this->mandatorId;
    }

    /**
     * @param int $mandatorId
     *
     * @return UserSession
     */
    public function setMandatorId($mandatorId)
    {
        $this->mandatorId = $mandatorId;

        return $this;
    }

    /**
     * @return string
     */
    public function getSessionId()
    {
        return $this->sessionId;
    }

    /**
     * @param int $sessionId
     *
     * @return UserSession
     */
    public function setSessionId($sessionId)
    {
        $this->sessionId = $sessionId;

        return $this;
    }

    /**
     * @return int
     */
    public function getUserId()
    {
        return $this->userId;
    }

    /**
     * @param int $userId
     *
     * @return UserSession
     */
    public function setUserId($userId)
    {
        $this->userId = $userId;

        return $this;
    }

    /**
     * Get system database component
     *
     * @return \isys_component_database|mixed|object
     * @throws \Exception
     */
    protected function getSystemDatabase()
    {
        return \isys_application::instance()->container->get('database_system');
    }

    /**
     * @throws \isys_exception_api
     * @throws \Exception
     */
    public function login()
    {
        // Check whether requirements are met
        if (empty($this->getUsername()) || empty($this->getPassword()) || empty($this->getApiKey())) {
            throw new \Exception('Please provide valid username, password and apikey information');
        }

        // Initialize session related data
        $this->initialize();

        // Check whether presistence is wanted and session exists already
        if ($this->isPersistent() && !empty($this->getSessionId())) {
            // Login via sessionId
            return $this->getSessionComponent()
                ->apikey_login($this->getApiKey(), null, $this->getSessionId());
        }

        // Reqular login via api key, username and password
        if ($this->getSessionComponent()
            ->apikey_login($this->getApiKey(), ['username' => $this->getUsername(), 'password' => $this->getPassword()])) {
            // Set sessionId and userId
            $this->setSessionId($this->getSessionComponent()->get_session_id())
                ->setUserId($this->getSessionComponent()->get_user_id());

            return true;
        }

        return false;
    }

    /**
     * Initialize login procedure
     *
     * @throws \Exception
     * @return bool
     */
    protected function initialize()
    {
        // Get mandator by api key
        $mandatorId = $this->getMandatorIdByApiKey($this->getApiKey());

        // Connect mandator now
        $this->getSessionComponent()
            ->connect_mandator($mandatorId);

        // Get user by credentials now
        $userId = $this->getUserIdByCredentials($this->getUsername(), $this->getPassword());

        // Check whether persistent session should be used
        if ($this->isPersistent()) {
            try {
                // Get session for user
                $this->setSessionId($this->getSessionByUser($userId));
            } catch (\Exception $e) {

            }
        }

        // Set mandatorId and userId
        $this->setMandatorId($mandatorId)
            ->setUserId($userId);

        return true;
    }

    /**
     * Get session by userId
     *
     * @param int $userId
     *
     * @return string
     * @throws \Exception
     */
    public function getSessionByUser($userId)
    {
        $sql = 'SELECT * FROM isys_user_session WHERE isys_user_session__isys_obj__id = ' . $userId;

        $resource = $GLOBALS['g_comp_database']->query($sql);

        if ($GLOBALS['g_comp_database']->num_rows($resource)) {
            return $GLOBALS['g_comp_database']->fetch_row_assoc($resource)['isys_user_session__php_sid'];
        }

        throw new \Exception('Unable to find session for user.');
    }

    /**
     * Get mandatorId by apiKey
     *
     * @param string $apiKey
     *
     * @return int
     * @throws \Exception
     */
    public function getMandatorIdByApiKey($apiKey)
    {
        $sql = 'SELECT isys_mandator__id FROM isys_mandator WHERE isys_mandator__apikey = \'' . $apiKey . '\' AND isys_mandator__active = 1';

        $resource = $this->getSystemDatabase()
            ->query($sql);

        if ($this->getSystemDatabase()
            ->num_rows($resource)) {
            return (int)$this->getSystemDatabase()
                ->fetch_row_assoc($resource)['isys_mandator__id'];
        }

        throw new \Exception('Unable to retrieve mandator');
    }

    /**
     * Get userId by credentials
     *
     * @param string $username
     * @param string $password
     *
     * @return int
     * @throws \Exception
     */
    public function getUserIdByCredentials($username, $password)
    {
        $sql = 'SELECT * FROM isys_obj object
                INNER JOIN isys_cats_person_list person ON object.isys_obj__id = person.isys_cats_person_list__isys_obj__id
                WHERE BINARY LOWER(person.isys_cats_person_list__title) = LOWER(\'' . $this->getUsername() . '\') AND
                      isys_cats_person_list__user_pass = md5(\'' . $this->getPassword() . '\')
        ';

        $resource = $GLOBALS['g_comp_database']->query($sql);

        if ($GLOBALS['g_comp_database']->num_rows($resource) == 1) {
            return (int)$GLOBALS['g_comp_database']->fetch_row_assoc($resource)['isys_obj__id'];
        }

        throw new \Exception('Unable to get userId by credentials');
    }

    /**
     * Destruction procedure
     */
    public function __destruct()
    {
        if (!$this->isPersistent()) {
            $this->getSessionComponent()
                ->logout();
        } else {
            try {
                $this->refreshSession();
            } catch (\Exception $e) {
                // We do not need to handle it!
            }
        }
    }

    /**
     * UserSession constructor.
     *
     * @param string                 $username
     * @param string                 $password
     * @param string                 $apiKey
     * @param bool                   $persistent
     * @param isys_component_session $sessionComponent
     */
    public function __construct($username, $password, $apiKey, $persistent, isys_component_session $sessionComponent)
    {
        $this->setUsername($username)
            ->setPassword($password)
            ->setApiKey($apiKey)
            ->setPersistent($persistent)
            ->setSessionComponent($sessionComponent);
    }

    /**
     * Logout
     */
    public function logout()
    {
        $this->__destruct();
    }
}